#import "ChipmunkMultiGrab.h"


// A constraint subclass that tracks a grab point
@interface ChipmunkGrab : NSObject<ChipmunkObject> {
	NSArray *_chipmunkObjects;
	
	cpVect _pos;
	cpFloat _smoothing;
}

@property(nonatomic, readonly) NSArray *chipmunkObjects;
@property(nonatomic, assign) cpVect pos;

@end


@implementation ChipmunkGrab

@synthesize pos = _pos;
@synthesize chipmunkObjects = _chipmunkObjects;

static void 
GrabPreSolve(cpConstraint *constraint, cpSpace *space)
{
	cpBody *grabBody = cpConstraintGetA(constraint);
	ChipmunkGrab *grab = [ChipmunkConstraint constraintFromCPConstraint:constraint].data;
	cpFloat dt = cpSpaceGetCurrentTimeStep(space);
	cpFloat coef = cpfpow(grab->_smoothing, dt);
	
	// Smooth out the mouse position.
	cpVect pos = cpvlerp(grab->_pos, cpBodyGetPos(grabBody), coef);
	cpBodySetVel(grabBody, cpvmult(cpvsub(pos, cpBodyGetPos(grabBody)), 1.0/dt));
	cpBodySetPos(grabBody, pos);
}

// Body will be nil if no object was grabbed.
-(id)initWithMultiGrab:(ChipmunkMultiGrab *)multiGrab pos:(cpVect)pos nearest:(cpVect)nearest body:(ChipmunkBody *)body chipmunkObjects:(NSArray *)chipmunkObjects;
{
	ChipmunkBody *grabBody = [ChipmunkBody bodyWithMass:INFINITY andMoment:INFINITY];
	grabBody.pos = pos;
	
	if((self = [super init])){
		_pos = pos;
		_smoothing = multiGrab.smoothing;
		
		if(body){
			ChipmunkPivotJoint *pivot = [ChipmunkPivotJoint pivotJointWithBodyA:grabBody bodyB:body anchr1:cpvzero anchr2:[body world2local:nearest]];
			pivot.maxForce = multiGrab.grabForce;
			pivot.data = self;
			cpConstraintSetPreSolveFunc(pivot.constraint, GrabPreSolve);
			chipmunkObjects = [chipmunkObjects arrayByAddingObject:pivot];
			
			cpFloat frictionForce = multiGrab.grabFriction;
			if(frictionForce > 0.0){
				ChipmunkPivotJoint *friction = [ChipmunkPivotJoint pivotJointWithBodyA:grabBody bodyB:body anchr1:cpvzero anchr2:[body world2local:nearest]];
				friction.maxForce = frictionForce;
				friction.maxBias = 0.0;
				chipmunkObjects = [chipmunkObjects arrayByAddingObject:friction];
			}
			
			cpFloat rotaryFriction = multiGrab.grabRotaryFriction;
			if(rotaryFriction > 0.0){
				ChipmunkGearJoint *friction = [ChipmunkGearJoint gearJointWithBodyA:grabBody bodyB:body phase:0.0 ratio:1.0];
				friction.maxForce = rotaryFriction;
				friction.maxBias = 0.0;
				chipmunkObjects = [chipmunkObjects arrayByAddingObject:friction];
			}
			
			_chipmunkObjects = [chipmunkObjects retain];
		}
	}
	
	return self;
}

-(void)dealloc
{
	[_chipmunkObjects release]; _chipmunkObjects = nil;
	[super dealloc];
}

@end


@implementation ChipmunkMultiGrab

@synthesize grabForce = _grabForce;
@synthesize smoothing = _smoothing;

@synthesize layers = _layers, group = _group;
@synthesize grabFilter = _grabFilter;

@synthesize grabFriction = _grabFriction, grabRotaryFriction = _grabRotaryFriction;
@synthesize grabRadius = _grabRadius;

@synthesize pullMode = _pullMode, pushMode = _pushMode;

@synthesize pushMass = _pushMass;
@synthesize pushFriction = _pushFriction, pushElasticity = _pushElasticity;
@synthesize pushCollisionType = _pushCollisionType;

-(id)initForSpace:(ChipmunkSpace *)space withSmoothing:(cpFloat)smoothing withGrabForce:(cpFloat)grabForce;
{
	if((self = [super init])){
		_space = [space retain];
		_grabs = [[NSMutableArray alloc] init];
		
		_smoothing = smoothing;
		_grabForce = grabForce;
		
		_layers = CP_ALL_LAYERS;
		_group = CP_NO_GROUP;
		
		_grabFilter = ^(ChipmunkShape *shape){return (bool)TRUE;};
		_pullMode = TRUE;
		_pushMode = FALSE;
	}
	
	return self;
}

-(void)dealloc
{
	[_space release];
	[_grabs release];
	
	[super dealloc];
}

-(BOOL)beginLocation:(cpVect)pos;
{
	__block cpFloat min = INFINITY;
	__block cpVect nearest = pos;
	__block ChipmunkBody *body = nil;
	
	if(_pullMode){
		cpSpaceNearestPointQuery_b(_space.space, pos, _grabRadius, _layers, _group, ^(cpShape *c_shape, cpFloat dist, cpVect point){
			if(dist < min && cpBodyGetMass(cpShapeGetBody(c_shape)) != INFINITY){
				ChipmunkShape *shape = [ChipmunkShape shapeFromCPShape:c_shape];
				if(_grabFilter(shape)){
					min = dist;
					nearest = (dist > 0.0 ? point : pos);
					body = shape.body;
				}
			}
		});
	}
	
	NSArray *chipmunkObjects = [NSArray array];
	if(!body && _pushMode){
		body = [ChipmunkBody bodyWithMass:_pushMass andMoment:INFINITY];
		body.pos = pos;
		
		ChipmunkShape *pushShape = [ChipmunkCircleShape circleWithBody:body radius:_grabRadius offset:cpvzero];
		pushShape.friction = _pushFriction;
		pushShape.elasticity = _pushElasticity;
		pushShape.layers = _layers;
		pushShape.group = _group;
		pushShape.collisionType = _pushCollisionType;
		
		chipmunkObjects = [NSArray arrayWithObjects:body, pushShape, nil];
	}
	
	ChipmunkGrab *grab = [[ChipmunkGrab alloc] initWithMultiGrab:self pos:pos nearest:nearest body:body chipmunkObjects:chipmunkObjects];
	
	[_grabs addObject:grab];
	[_space add:grab];
	[grab release];
	
	return (body != nil);
}

static ChipmunkGrab *
BestGrab(NSArray *grabs, cpVect pos)
{
	ChipmunkGrab *match = nil;
	cpFloat best = INFINITY;
	
	for(ChipmunkGrab *grab in grabs){
		cpFloat dist = cpvdistsq(pos, grab.pos);
		if(dist < best){
			match = grab;
			best = dist;
		}
	}
	
	return match;
}

-(void)updateLocation:(cpVect)pos;
{
	BestGrab(_grabs, pos).pos = pos;
}

-(void)endLocation:(cpVect)pos;
{
	cpAssertHard([_grabs count] != 0, "Grab set is already empty!");
	ChipmunkGrab *grab = BestGrab(_grabs, pos);
	
	[_space remove:grab];
	[_grabs removeObject:grab];
}

@end
